/* Translates "Tnop", "Tsession", etc. into the 9P protocol. */
/* Ben Lin
 * 
 *	20/3/98
 * 		Started coding. Only had nop and session.
 * 	24/3/98
 * 		Added everything except create, read, write.
 * 	27/3/98
 * 		Added create, read, write.
 */

#include <stdio.h>
#include "unix9p.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#define BACKLOG 5

#define num16() scanf("%d", &i); put16(i)
#define num32() scanf("%d", &i); put32(i)
#define num64() scanf("%lld", &lli); put64(lli)
#define string(n) scanf("%s", str); putfield(str, n);

int sfd;

/* output a byte */

void put1(char ch) {
	char c;

	c = ch;
	write(sfd, &c, 1);
}

/* output a 16-bit little-endian unsigned integer */

void put16(int i) {
	/* little-endian */
	put1((char) i);	/* high-bits get truncated */
	put1((char) (i >> 8));
}

/* put a 32-bit little-endian unsigned integer */

void put32(int i) {
	put1((char) i);
	put1((char) (i = i >> 8));
	put1((char) (i = i >> 8));
	put1((char) (i >> 8));
}

/* put a 64-bit little-endian unsigned integer */

void put64(long long int i) {
	put1((char) i);
	put1((char) (i = i >> 8));
	put1((char) (i = i >> 8));
	put1((char) (i = i >> 8));
	put1((char) (i = i >> 8));
	put1((char) (i = i >> 8));
	put1((char) (i = i >> 8));
	put1((char) (i >> 8));
}

/* gets a field of n bytes, returns 0 if successful, 1 otherwise */

char getfield(char *s, int n) {
	int i;

	for(i=0; i<n; i++) {
		s[i] = getchar();
	}

	return(0);
}

/* puts a field of n bytes */

void putfield(char *s, int n) {
	int i;

	write(sfd, s, strlen(s));
	put1('\0');
	n = n - strlen(s);
	for (i=1; i<n; i++) {
		put1('\0');
	}
}

main(int argc, char *argv[]) {
	char str[100];
	int i, tag;
	long long int lli;
	struct sockaddr_in sock9P;
	
	printf("Getting socket...");
	sfd = socket(AF_INET, SOCK_STREAM, 0);

	if (sfd == -1) {
		fprintf(stderr, "Couldn't get a socket\n");
		exit(1);
	}

	printf("done\n");

	sock9P.sin_family = AF_INET;
	sock9P.sin_port = htons(PORT_9P);
	sock9P.sin_addr.s_addr = inet_addr("127.0.0.1");
	bzero(&(sock9P.sin_zero), 8);

	if (connect(sfd, (struct sockaddr *)&sock9P, sizeof(struct sockaddr)) != 0) {
		fprintf(stderr, "Couldn't connect the socket\n");
		exit(1);
	}
	printf("Connected\n");

	scanf("%s", &str);
	while (!feof(stdin)) {

		if ((str[0] | 32) == 'T') {
			i=0;
			while (str[i] = str[++i]);
		}
		if (!strcmp("nop", str)) {
			put1(Tnop);
			put16(NOTAG);
		} else if (!strcmp("session", str)) {
			put1(Tsession);
			put16(NOTAG);
			for (i=0; i<8; i++) {
				put1('*');
			}
		} else if (!strcmp("attach", str)) {
			put1(Tattach);
			num16(); /* tag */
			num16(); /* fid */
			string(NAMELEN); /* uid */
			string(NAMELEN); /* aname */
			/*
			getfield(str, TICKETLEN + AUTHENTLEN);
			putfield(str, TICKETLEN + AUTHENTLEN);
			*/
			for (i=0; i < TICKETLEN + AUTHENTLEN; i++) {
				put1('*');
			}
		} else if (!strcmp("clone", str)) {
			put1(Tclone);
			num16(); /* tag */
			num16(); /* fid */
			num16(); /* newfid */

		} else if (!strcmp("clunk", str)) {
			put1(Tclunk);
			num16(); /* tag */
			num16(); /* fid */

		} else if (!strcmp("clwalk", str)) {
			put1(Tclwalk);
			num16(); /* tag */
			num16(); /* fid */
			num16(); /* newfid */
			string(NAMELEN);	/* name */

		} else if (!strcmp("flush", str)) {
			put1(Tflush);
			num16(); /* tag */
			num16(); /* oldtag */

		} else if (!strcmp("open", str)) {
			put1(Topen);
			num16(); /* tag */
			num16(); /* fid */
			scanf("%d", &i); put1(i); /* mode */

		} else if (!strcmp("create", str)) {
			put1(Tcreate);
			num16(); /* tag */
			num16(); /* fid */
			string(NAMELEN);	/* name */
			num32(); /* perm */
			scanf("%d", &i); put1(i); /* mode */

		} else if (!strcmp("read", str)) {
			put1(Tread);
			num16(); /* tag */
			num16(); /* fid */
			num64(); /* offset */
			num16(); /* count */

		} else if (!strcmp("write", str)) {
			put1(Twrite);
			num16(); /* tag */
			num16(); /* fid */
			num64(); /* offset */
			num16(); /* count */
			getchar(); /* pad */
			string(i);

		} else if (!strcmp("remove", str)) {
			put1(Tremove);
			num16(); /* tag */
			num16(); /* fid */

		} else if (!strcmp("stat", str)) {
			put1(Tstat);
			num16(); /* tag */
			num16(); /* fid */

		} else if (!strcmp("wstat", str)) {
			put1(Twstat);
			num16(); /* tag */
			num16(); /* fid */
			string(DIRLEN) /* stat */

		} else if (!strcmp("walk", str)) {
			put1(Twalk);
			num16(); /* tag */
			num16(); /* fid */
			string(28);	/* name */
		}

		scanf("%s", &str);
	}
	put1(Tmax);
	close(sfd);

	return(0);
}
