/* 9P/Styx File Server */
/* Ben Lin */

/* No authentication is performed. This deviates slightly from the protocol
because it accepts filenames exactly 28 characters long (so they aren't
null-terminated) */

/*
 *  18/3/98
 *		Started coding. Just had the skeleton.
 *  20/3/98
 * 		Added nop, session.
 *  22/3/98
 * 		Replaced case statements with an array of pointers to functions.
 * 		Made tag a global variable.
 * 	23/3/98
 * 		Changed gettag, puttag to get16 and put16. Added linked list stuff.
 * 		Started attach
 * 	24/3/98
 * 		attach appears to be working. Added clone.
 * 	25/3/98
 * 		Fixed minor bugs - I had a ! instead of a ~. Checked for errors
 * 		AFTER getting fields. Added clunk, flush, walk, GOODTAG.
 * 	28/3/98
 * 		Added open, read, stat.
 * 	30/3/98
 * 		Added write, remove, create.
 *  9/5/98
 *		Added socket code
 *  16/5/98
 *		Added getpacket, putpacket, made tag local again.
 *	17/5/98
 *		Fixed up read for directories, still can't do group and user names
 *  18/5/98
 *		Added group and user names, tweaked a few things. Fixed bug in
 *		getpacket (it didn't handle 0 length fields properly)
 *	11/6/98
 *		Over the past few weeks: added fork for device I/O, and (limited)
 *	support for append-only. Exclusive use will probably require
 *	another linked-list.
 */

/* TODO:
 * Honour exclusive use, fix group id problem
 * use strerror(errno) in some places(?):
 * errmsg(tag, strerror(errno));
 */

#define VERBOSE

#define BACKLOG 3
#define HASHSIZE 107
#define HASHCONST 53

#define GOODTAG(message) if (tag == NOTAG) { \
		errmsg(tag, "message: tag should not be NOTAG (0xFFFF)"); \
		return; \
	} \
	if (pendretrieve(tag)) { \
		errmsg(tag, "message: tag in use"); \
		return; \
	}

#define HASH(n) (n * HASHCONST % HASHSIZE)

#define getpacket(args...) fgetpacket(nfd, ## args)
#define putpacket(args...) fputpacket(nfd, ## args)

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <utime.h>
#include <pwd.h>
#include <grp.h>
#include <dirent.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include "unix9p.h"

typedef unsigned int fidtype;

/* Keeps track of all the fids */
typedef struct fidliststruct {
	fidtype fid;
	char name[28];
	char *truename;
	char open;
	char mode;
	int qidpath;
	int qidver;
	int fd;
	struct fidliststruct *next;
	int childfd;	/* if not -1, use this to talk to child */
	char appendonly;
	char exclusiveuse;
} fidelem;

/* Keeps track of pending reads and writes */
typedef struct pendliststruct {
	int tag;
	int fid;
	pid_t pid;		/* PID of the child */
	struct pendliststruct *next;
} pendelem;

typedef struct {
	fidelem *list;	/* linked list of fid's */ 
	fidelem *end;
} fidhtelem;

/* global variables */

int ocount = 0;	/* keeps track of open files */
int nfd;	/* new file descriptor - needed for the accept call */
fidtype fid;
fidhtelem fidht[HASHSIZE];

pendelem *pendlist;	/* linked list of pending tags */ 
pendelem *pendlistend;

char data[MAXFDATA];
int parentfd[2];	/* pipe that kids talk to parent through */

/* function prototypes */

void handlenop();
void handlesession();
void handleflush();
void handleattach();
void handleclone();
void handleclwalk();
void handlewalk();
void handleopen();
void handlecreate();
void handleread();
void handlewrite();
void handleclunk();
void handleremove();
void handlestat();
void handlewstat();
void errmsg(int, char *);

/* Safe version of malloc */
/* Exits with error message if out of memory */

void *salloc(size_t size) {
	void *ptr;

	if (!(ptr = malloc(size))) {
		fprintf(stderr, "Out of memory\n");
		exit(200);
	}
	
	return(ptr);
}

/* puts a field of n bytes */

void putfield(unsigned char *s, int n) {
	int i;

	for (i=0; i<n; i++) {
		fprintf(stderr, " %.2x", s[i]);
	}
}

void put4(char *s, int n) {
	int i;
	
	i = 0;
	s[i++] = (char) n;
	s[i++] = (char) (n = n >> 8);
	s[i++] = (char) (n = n >> 8);
	s[i++] = (char) (n >> 8);
}

int fputpacket(int fd, char *fmt, ...) {
	va_list ap;
	static char packet[MAXFDATA + 16];
	char ch, *s;
	long long int ln;
	int n;
	int i, j;

	va_start(ap, fmt);
	i = 0;
	while(*fmt) {
		switch(*fmt++) {
			case '1':	/* one byte */
				ch = va_arg(ap, char);
				packet[i++] = ch;
				break;

			case '2':
				n = va_arg(ap, int);
				packet[i++] = (char) n;	/* high-bits get truncated */
				packet[i++] = (char) (n >> 8);
				break;
			case '4':
				n = va_arg(ap, int);
				packet[i++] = (char) n;
				packet[i++] = (char) (n = n >> 8);
				packet[i++] = (char) (n = n >> 8);
				packet[i++] = (char) (n >> 8);
				break;
			case '8':
				ln = va_arg(ap, long int);
				packet[i++] = (char) ln;
				packet[i++] = (char) (ln = ln >> 8);
				packet[i++] = (char) (ln = ln >> 8);
				packet[i++] = (char) (ln = ln >> 8);
				packet[i++] = (char) (ln = ln >> 8);
				packet[i++] = (char) (ln = ln >> 8);
				packet[i++] = (char) (ln = ln >> 8);
				packet[i++] = (char) (ln >> 8);
				break;

			case 's':
				s = va_arg(ap, char *);
				n = va_arg(ap, int);
				for (j=0; j<n; j++) {
					packet[i++] = s[j];
				}
				break;

			default:
				fprintf(stderr, "9ps: BUG! Bad char in format string\n");
				break;
		}
	}

	va_end(ap);

	if (write(fd, packet, i) != i) {
		perror("write");
		exit(10);
	}
	fprintf(stderr, "%d bytes written\n", i);
	//putfield(packet, i);

	return(0);
}

int fgetpacket(int fd, char *fmt, ...) {
	va_list ap;
	static unsigned char packet[MAXFDATA + 16];
	int length;
	int i, j;
	int *ip;
	long long int *lp;
	char *cp;

	va_start(ap, fmt);
	length = 0;
	i = 0;
	while(fmt[i]) {
		switch(fmt[i++]) {
			case '1':	/* one byte */
				va_arg(ap, char);
				length++;
				break;

			case '2':
				va_arg(ap, char);
				length+=2;
				break;

			case '4':
				va_arg(ap, char);
				length+=4;
				break;

			case '8':
				va_arg(ap, char);
				length+=8;
				break;

			case 's':
				va_arg(ap, char);
				length+=va_arg(ap, int);
				break;

			default:
				fprintf(stderr, "9ps: BUG! Bad char in format string\n");
				break;
		}
	}
	
	i = 0;
	while (length > 0) {
	fprintf(stderr, "reading...\n");
		j = read(fd, &packet[i], length);
	fprintf(stderr, "(%d bytes)\n", j);
		if (j == -1) {
			errmsg(NOTAG, "Error reading from port");
			fprintf(stderr, "9ps: error reading from port");
			/* TODO: stop trying after a while */
		} else if (j == 0) {
			fprintf(stderr, "connection lost.\n");
			exit(10);
		} else {
			i += j;
			length -= j;
		}
	}

	i = 0;
	va_start(ap, fmt);

	while(*fmt) {
		switch(*fmt++) {
			case '1':	/* one byte */
				cp = va_arg(ap, char *);
				*cp = packet[i++];
				break;

			case '2':
				ip = va_arg(ap, int *);
				*ip = packet[i++]; 
				*ip |= packet[i++] << 8;
				break;

			case '4':
				ip = va_arg(ap, int *);
				*ip = packet[i++]; 
				*ip |= packet[i++] << 8;
				*ip |= packet[i++] << 16;
				*ip |= packet[i++] << 24;
				break;

			case '8':
				lp = va_arg(ap, long long int *);
				*lp = packet[i++];
				for (j=1; j<8; j++) {
					*lp |= ((long long int) packet[i++]) << (8*j);
				}
				break;

			case 's':
				cp = va_arg(ap, char *);
				length = va_arg(ap, int);
				for (j=0; j<length; j++) {
					cp[j] = packet[i++];
				}

				break;

			default:
				fprintf(stderr, "9ps: BUG! Bad char in format string\n");
				break;
		}
	}
	va_end(ap);

	return(0);
}

int qidpath(struct stat stbuf) {
	int i;
	i = stbuf.st_ino;

	if (S_ISDIR(stbuf.st_mode)) {
		i |= DIRMASK;
	} else {
		i &= ~DIRMASK;
	}

	return(i);

}

int qidver(struct stat stbuf) {
   return(0);
}

/* Replies with an error message */

void errmsg(int tag, char *s) {
	char ename[ERRLEN];

	if (strlen(s) > ERRLEN) {
		fprintf(stderr, "BUG: Error string too long\n");
		exit(1);
	}

	strcpy(ename, s);

#ifdef VERBOSE
	fprintf(stderr, "<Sending error message: %s>", ename);
#endif

	putpacket("12s", Rerror, tag, ename, ERRLEN);
}

/* retrieve element from fid table */

fidelem *fidretrieve(fidtype fid) {
	fidelem *f;

	f = fidht[HASH(fid)].list->next;

	for(;;) {
		if (!f) {
			return(NULL);
		}

		if (f->fid == fid) {
			return(f);
		}

		f = f->next;
	}
}

/* delete element from fid table */

fidelem *fiddelete(fidtype fid) {
	fidelem *temp, *f;
	int n;

	n = HASH(fid);
	f = fidht[n].list;
	while (f->next) {
		if (f->next->fid == fid) {
			temp = f->next;
			f->next = f->next->next;

			/* Don't forget this special case: */
			if (!f->next) fidht[n].end = f;

			return(temp);
		}

		f = f->next;
	}

	return(NULL);
}

/* make new entry */

fidelem *fidnew(fidtype fid, char *name, char *truename, int qp, int qv) {
	fidelem *e;
	int l;

	e = (fidelem *) salloc(sizeof(fidelem));

	e->fid = fid;
	strcpy(e->name, name);
	e->truename = (char *) salloc((l = strlen(truename)) + 1);
	strcpy(e->truename, truename);
	e->truename[l] = '\0';
	e->open = 0;
	e->qidpath = qp;
	e->qidver = qv;
	e->childfd = -1;
	e->appendonly = 0;
	e->exclusiveuse = 0;

	return(e);
}

/* insert entry into hash table */

void fidinsert(fidelem *elem) {
	int n;

	n = HASH(elem->fid);
	fidht[n].end->next = elem;
	fidht[n].end = elem;
	elem->next = NULL;
}

/* make new pending tag entry */

pendelem *pendnew(int tag, int fid, pid_t pid) {
	pendelem *p;
	
	p = (pendelem *) salloc(sizeof(pendelem));
	p->tag = tag;
	p->fid = fid;
	p->pid = pid;

	return(p);
}

/* add pending tag to the list. Doesn't check for duplicates */

void pendadd(pendelem *elem) {
	pendlistend->next = elem;
	pendlistend = elem;
	elem->next = NULL;
}
	
/* retrieve element from fid table */

pendelem *pendretrieve(int tag) {
	pendelem *f;

	f = pendlist->next;

	for(;;) {
		if (!f) {
			return(NULL);
		}

		if (f->tag == tag) {
			return(f);
		}

		f = f->next;
	}
}

/* delete element from pendlist */

pendelem *penddelete(int tag) {
	pendelem *temp, *f;

	f = pendlist;
	while (f->next) {
		if (f->next->tag == tag) {
			temp = f->next;
			f->next = f->next->next;

			/* Don't forget this special case: */
			if (!f->next) pendlistend = f;

			return(temp);
		}

		f = f->next;
	}

	return(NULL);
}

/* print pendlist */

void pendprint(void) {
	pendelem *f;

	f = pendlist->next;
	while (f) {
		fprintf(stderr, "tag: %d, fid: %d\n", f->tag, f->fid);
		f = f->next;
	}
}




/* MAIN */

main(int argc, char *argv[]) {
	unsigned char ch;
	int i;
	fd_set rfds;
	void (*jumptable[Tmax])();
	int sfd, maxfd;
	struct sockaddr_in sock9P;
	struct sockaddr_in sockclient;
	int sin_size;
	int tag;
	fidelem *felem;
	pendelem *pelem;
	int length;
	char message[8 + MAXFDATA];
#ifdef VERBOSE
	char tname[Tmax][10];
#endif


#ifdef VERBOSE
	fprintf(stderr, "VERBOSE defined - showing all messages\n");
#endif
	/* ino_t > 4 bytes => qid's may not be unique */

	/* sentinel at head of each list */
	for(i=0; i<HASHSIZE; i++) {
		fidht[i].list = (fidelem *) salloc(sizeof(fidelem));
		fidht[i].end = fidht[i].list;
		fidht[i].list->next = NULL;
	}

	/* sentinel at head of list */
	pendlist = (pendelem *) salloc(sizeof(pendelem));
	pendlistend = pendlist;
	pendlist->next = NULL;

	for (i=0; i<Tmax; i++) {
		jumptable[i] = NULL;
#ifdef VERBOSE
		tname[i][0] = 0;
#endif
	}

#ifdef VERBOSE
	strcpy(tname[Tnop], "Tnop");
	strcpy(tname[Rnop], "Rnop");
	strcpy(tname[Terror], "Terror");
	strcpy(tname[Rerror], "Rerror");
	strcpy(tname[Tflush], "Tflush");
	strcpy(tname[Rflush], "Rflush");
	strcpy(tname[Tclone], "Tclone");
	strcpy(tname[Rclone], "Rclone");
	strcpy(tname[Twalk], "Twalk");
	strcpy(tname[Rwalk], "Rwalk");
	strcpy(tname[Topen], "Topen");
	strcpy(tname[Ropen], "Ropen");
	strcpy(tname[Tcreate], "Tcreate");
	strcpy(tname[Rcreate], "Rcreate");
	strcpy(tname[Tread], "Tread");
	strcpy(tname[Rread], "Rread");
	strcpy(tname[Twrite], "Twrite");
	strcpy(tname[Rwrite], "Rwrite");
	strcpy(tname[Tclunk], "Tclunk");
	strcpy(tname[Rclunk], "Rclunk");
	strcpy(tname[Tremove], "Tremove");
	strcpy(tname[Rremove], "Rremove");
	strcpy(tname[Tstat], "Tstat");
	strcpy(tname[Rstat], "Rstat");
	strcpy(tname[Twstat], "Twstat");
	strcpy(tname[Rwstat], "Rwstat");
	strcpy(tname[Tclwalk], "Tclwalk");
	strcpy(tname[Rclwalk], "Rclwalk");
	strcpy(tname[Tauth], "Tauth");
	strcpy(tname[Rauth], "Rauth");
	strcpy(tname[Tsession], "Tsession");
	strcpy(tname[Rsession], "Rsession");
	strcpy(tname[Tattach], "Tattach");
	strcpy(tname[Rattach], "Rattach");
#endif

	jumptable[Tnop] = handlenop;
	jumptable[Tsession] = handlesession;
	jumptable[Tattach] = handleattach;
	jumptable[Tclone] = handleclone;
	jumptable[Tclunk] = handleclunk;
	jumptable[Tflush] = handleflush;
	jumptable[Topen] = handleopen;
	jumptable[Tcreate] = handlecreate;
	jumptable[Tread] = handleread;
	jumptable[Twrite] = handlewrite;
	jumptable[Tremove] = handleremove;
	jumptable[Tstat] = handlestat;
	jumptable[Twstat] = handlewstat;
	jumptable[Tclwalk] = handleclwalk;
	jumptable[Twalk] = handlewalk;
	
	fprintf(stderr, "Setting up pipe...");
	if (pipe(parentfd)) {
		perror("pipe");
		exit(1);
	}
	fprintf(stderr, "done\n");
	fprintf(stderr, "Testing pipe %d...", parentfd[1]);
	write(parentfd[1], message, 40);
	if (read(parentfd[0], message, 40) != 40) {
		fprintf(stderr, "pipe failed\n");
		exit(1);
	}
	fprintf(stderr, "done\n");

	fprintf(stderr, "Getting socket...");
	sfd = socket(AF_INET, SOCK_STREAM, 0);

	if (sfd == -1) {
		perror("socket");
		exit(1);
	}

	fprintf(stderr, "done\n");

	sock9P.sin_family = AF_INET;
	sock9P.sin_port = htons(PORT_9P);
	sock9P.sin_addr.s_addr = INADDR_ANY;
	bzero(&(sock9P.sin_zero), 8);

	fprintf(stderr, "Binding port %d...", PORT_9P);

	if (bind(sfd, (struct sockaddr *)&sock9P, sizeof(struct sockaddr)) != 0) {
		perror("bind");
		exit(1);
	}
	fprintf(stderr, "done\nListening:\n");

	if (listen(sfd, BACKLOG) == -1) {
		perror("listen");
		exit(1);
	}

	sin_size = sizeof(struct sockaddr_in);

	fprintf(stderr, "Accepting...\n");
	nfd = accept(sfd, (struct sockaddr *) &sockclient, &sin_size);

	FD_ZERO(&rfds);
	if (nfd > parentfd[0])
		maxfd = nfd + 1;
	else
		maxfd = parentfd[0] + 1;

	for(;;) {

		fprintf(stderr, "selecting...\n");
		FD_SET(nfd, &rfds);
		FD_SET(parentfd[0], &rfds);
		if (select(maxfd, &rfds, NULL, NULL, NULL) == -1) {
			perror("select");
			break;
		}
		
		if (FD_ISSET(parentfd[0], &rfds)) {
		/* a child's trying to tell us something */

			fprintf(stderr, "Message from child\n");
			fgetpacket(parentfd[0], "22", &tag, &length);
			fgetpacket(parentfd[0], "s", message, length);
			pelem = penddelete(tag);
			felem = fidretrieve(pelem->fid);
			putpacket("s", message, length);
			close(felem->childfd);
			felem->childfd = -1;
			waitpid(pelem->pid, NULL, 0);
			free(pelem);

		}
		
		if (FD_ISSET(nfd, &rfds)) {
			fprintf(stderr, "Message from socket\n");
			getpacket("1", &ch);

#ifdef VERBOSE
			if ((ch < Tmax) && tname[ch][0]) {
				fprintf(stderr, "<%s>", tname[ch]);
			} else {
				fprintf(stderr, "<type = %d>", ch);
			}
#endif

			if (ch >= Tmax || !jumptable[ch]) {
				fprintf(stderr, "Bad message type received\n");
				errmsg(NOTAG, "Bad message type");

				/* should also read a tag anyway? */

				/* give up and try reading another message */
				continue;
			}

			(*jumptable[ch])();
		}
	}
	close(parentfd[0]);
	close(parentfd[1]);

	close(nfd);

	close(sfd);

	return(0);
}

void getstat(char *name, struct stat stbuf, char *s) {
	int i;
	int mode;
	struct passwd *pw;
	struct group *gr;

	i = 0;

	mode = stbuf.st_mode;
	if (S_ISDIR(stbuf.st_mode)) {
		mode |= DIRMASK;
	} else {
		mode &= ~DIRMASK;
	}

	strcpy(&s[i], name);
	i += NAMELEN;
	
	if (pw = getpwuid(stbuf.st_uid)) {
		strcpy(&s[i], pw->pw_name);
	} else {
		sprintf(&s[i], "%d", stbuf.st_uid);
	}

	i += NAMELEN;
	
	if (gr = getgrgid(stbuf.st_gid)) {
		strcpy(&s[i], gr->gr_name);
	} else {
		sprintf(&s[i], "%d", stbuf.st_gid);
	}
	i += NAMELEN;

	put4(&s[i], qidpath(stbuf));
	i += 4;
	put4(&s[i], qidver(stbuf));
	i += 4;
	put4(&s[i], mode);
	i += 4;
	put4(&s[i], stbuf.st_atime);
	i += 4;
	put4(&s[i], stbuf.st_mtime);
	i += 4;
	put4(&s[i], (int) (stbuf.st_size & 0xFFFFFFFF));
	i += 4;
	/* TODO: change this back? */
	// I guess Linux files can't be as big as Plan 9 files
	//put4(&s[i], (int) (stbuf.st_size >> 32));
	put4(&s[i], 0);
	i += 4;
	/* type and dev, "for kernel use" */
	s[i++] = 0;
	s[i++] = 0;
	s[i++] = 0;
	s[i++] = 0;

}

void handlenop() {
	int tag;
	char string[6];

	//getpacket("2s", &tag, string, 5);
	getpacket("2", &tag);

	if (tag != NOTAG) {
		fprintf(stderr, "%x\n", tag);
		tag = NOTAG;
		errmsg(tag, "Tnop: tag should be NOTAG (0xFFFF)");
		return;
	}

	putpacket("12", Rnop, NOTAG);
}

void handlesession() {
	char chal[CHALLEN], authid[NAMELEN], authdom[DOMLEN];
	int tag;

	getpacket("2s", &tag, chal, CHALLEN);

	if (tag != NOTAG) {
		tag = NOTAG;
		errmsg(tag, "Tsession: tag should be NOTAG (0xFFFF)");
		return;
	}

	/* garbage for now */
	strcpy(chal, "12345678");
	strcpy(authid, "abcdefghijklmnopqrstuvwxyz12");
	strcpy(authdom, "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUV");

	putpacket("12sss", Rsession, NOTAG, chal, CHALLEN,
			authid, NAMELEN, authdom, DOMLEN);

}

void handleattach() {
	struct stat stbuf;
	struct passwd *pw;
	int qidpath;
	int qidver;
	unsigned int qpath, qver;
	char uid[NAMELEN], aname[NAMELEN + 1], ticket[TICKETLEN],
		auth[AUTHENTLEN], rauth[AUTHENTLEN];
	int tag;

	aname[NAMELEN] = '\0';	/* Just in case it's missing */

#ifdef INFERNO
	getpacket("22ss", &tag, &fid, uid, NAMELEN, aname, NAMELEN);
#else
	getpacket("22ssss", &tag, &fid, uid, NAMELEN, aname, NAMELEN,
			ticket, TICKETLEN, auth, AUTHENTLEN);
#endif

	GOODTAG(Tattach);

	if (fidretrieve(fid)) {
		errmsg(tag, "Tattach: fid already used");
		return;
	}

	/* default is root directory */
	//if (!strlen(aname))
		strcpy(aname, "/");

	if (pw = getpwnam(uid)) {
		/* TODO: fairly insecure */
		if (setuid(pw->pw_uid)) {
			errmsg(tag, "Tattach: setuid failed");
			return;
		}
	} else {
		errmsg(tag, "Tattach: No such user");
		return;
	}

	if (stat(aname, &stbuf)) {
		fprintf(stderr, "error opening %s\n", aname);
		errmsg(tag, "Tattach: Unable to open aname");
		return;
	}

	qidpath = stbuf.st_ino;

	if (S_ISDIR(stbuf.st_mode)) {
		qidpath |= DIRMASK;
	} else {
		qidpath &= ~DIRMASK;
	}

	qidver = 0;

	fidinsert(fidnew(fid, aname, aname, qidpath, qidver));

	/* rauth */
	strcpy(rauth, "1234567890ABC");

#ifdef VERBOSE
	fprintf(stderr, "<Outputting...");
#endif

#ifdef INFERNO
	putpacket("12244", Rattach, tag, fid, qidpath, qidver);
#else
	putpacket("12244s", Rattach, tag, fid, qidpath, qidver, rauth, AUTHENTLEN);
#endif


#ifdef VERBOSE
	fprintf(stderr, "done>\n");
#endif
}

void handleclone(void) {
	fidelem *e;
	fidtype newfid;
	int tag;

	getpacket("222", &tag, &fid, &newfid);

	GOODTAG(Tclone);

	e = fidretrieve(fid);
	if (!e) {
		errmsg(tag, "Tclone: No such fid");
		return;
	}

	if (fidretrieve(newfid)) {
		errmsg(tag, "Tclone: newfid already used");
		return;
	}

	fidinsert(fidnew(newfid, e->name, e->truename, e->qidpath, e->qidver));

#ifdef VERBOSE
	fprintf(stderr, "<cloned %d -> %d>", fid, newfid);
#endif
	putpacket("122", Rclone, tag, fid);
}

void handleclunk(void) {
	fidelem *e;
	int tag;

	getpacket("22", &tag, &fid);

	GOODTAG(Tclunk);

	e = fiddelete(fid);

	if (!e) {
		errmsg(tag, "Tclunk: No such fid");
		return;
	}

	if (e->childfd != -1) {
		errmsg(tag, "Tclunk: fid busy");
		return;
	}

	if (e->open) {
ocount--;
fprintf(stderr, "CLOSE ***** %d\n", ocount);
		if (close(e->fd)) {
			perror("close");
		}
		if (e->mode & ORCLOSE) {
#ifdef VERBOSE
			fprintf(stderr, "<remove %s on clunk>", e->truename);
#endif
			if (unlink(e->truename)) {
				errmsg(tag, "Tclunk: unable to remove file on clunk");
				return;
			}
		}
	}
	free(e);

	putpacket("122", Rclunk, tag, fid);
}

void handleflush(void) {
	int tag, oldtag;
	pendelem *p;
	fidelem *e;

	getpacket("22", &tag, &oldtag);

	GOODTAG(Tflush);
	if (p = penddelete(oldtag)) {
		e = fidretrieve(p->fid);
		fputpacket(e->childfd, "1", Tflush);
	
		close(e->childfd);
		e->childfd = -1;
		waitpid(p->pid, NULL, 0);
		free(p);
	}
	putpacket("12", Rflush, tag);
}

void handleopen(void) {
	char mode;
	int openmode;
	fidelem *e;
	struct stat stbuf;
	int tag;

	getpacket("221", &tag, &fid, &mode);

	GOODTAG(Topen);

#ifdef VERBOSE
	fprintf(stderr, "<opening fid %d>", fid);
#endif

	e = fidretrieve(fid);
	if (!e) {
		errmsg(tag, "Topen: No such fid");
		return;
	}

	if (e->childfd != -1) {
		errmsg(tag, "Topen: fid busy");
		return;
	}

	if (e->open) {
		errmsg(tag, "Topen: file already open");
		return;
	}

	stat(e->truename, &stbuf);

	switch(mode & 3) {
		case 0:
			openmode = O_RDONLY;
			break;
		case 1:
			if (S_ISDIR(stbuf.st_mode)) {
				errmsg(tag, "Topen: directories cannot be written to");
				return;
			}
			openmode = O_WRONLY;
			break;
		case 2:
			if (S_ISDIR(stbuf.st_mode)) {
				errmsg(tag, "Topen: directories cannot be written to");
				return;
			}
			openmode = O_RDWR;
			break;
		case 3:
			if (!(stbuf.st_mode & S_IEXEC)) {
				errmsg(tag, "Topen: execution of non-executable attempted");
				return;
			}
	}
	
	/* check for OTRUNC bit */
	if (mode & OTRUNC) {
		if (S_ISDIR(stbuf.st_mode)) {
			errmsg(tag, "Topen: directories cannot be truncated");
			return;
		}
		openmode |= O_TRUNC;
	}

	/* and ORCLOSE bit */
	if (mode & ORCLOSE) {
		if (S_ISDIR(stbuf.st_mode)) {
			errmsg(tag, "Topen: directories cannot be removed on close");
			return;
		}
	}

ocount++;
fprintf(stderr, "OPEN *****%d\n", ocount);
	e->fd = open(e->truename, openmode);

	if (e->fd < 0) {
		fprintf(stderr, "Error: %s\n", e->truename);
		perror("open");
		errmsg(tag, "Topen: error opening file");
		return;
	}

	e->open = 1;
	e->mode = mode;

	putpacket("12244", Ropen, tag, fid, e->qidpath, e->qidver);
}

void handlecreate(void) {
	fidelem *e;
	char name[NAMELEN];
	int perm;
	char mode;
	char *str;
	struct stat stbuf;
	struct group *gr;
	int l;
	mode_t creatmode;
	umode_t dirperm;
	gid_t dirgr;
	int tag;

	getpacket("22s41", &tag, &fid, name, NAMELEN, &perm, &mode);

	GOODTAG(Tcreate);

	if (!strcmp(name, ".") || !strcmp(name, "..")) {
		errmsg(tag, "Tcreate: illegal name");
		return;
	}

	e = fidretrieve(fid);
	if (!e) {
		errmsg(tag, "Tcreate: no such fid");
		return;
	}

	if (stat(e->truename, &stbuf)) {
		errmsg(tag, "Tcreate: stat failed");
		return;
	}

	if (!S_ISDIR(stbuf.st_mode)) {
		errmsg(tag, "Tcreate: fid is not a directory");
		return;
	}

	dirperm = stbuf.st_mode;
	dirgr = stbuf.st_gid;

	l = strlen(e->truename);
	str = (char *) salloc(l + NAMELEN + 1);
	strcpy(str, e->truename);
	if (strcmp(str, "/")) {
		str[l] = '/';
		l++;
	}

	strcpy(&str[l], name);

	if (!stat(str, &stbuf)) {
		errmsg(tag, "Tcreate: file already exists");
		return;
	}

	if (perm & AOMASK) {
		fprintf(stderr, "warning: append-only only while this file is open");
		perm = perm&(~AOMASK);
		e->appendonly = 1;
	}

	if (perm & EUMASK) {
		fprintf(stderr, "warning: exclusive-use not implemented at all");
		perm = perm&(~EUMASK);
		e->exclusiveuse = 1;
	}

	if (perm & DIRMASK) {
		creatmode = (perm&~0777) | (dirperm&perm&0777);
		if (mkdir(str, creatmode)) {
			errmsg(tag, "Tcreate: unable to make directory");
			return;
		}

		e->fd = open(str, O_RDONLY);
	} else {
		creatmode = (perm&(~0777|0111)) | dirperm&perm&0666;

		e->fd = creat(str, creatmode);
	}

ocount++;
fprintf(stderr, "OPEN *****%d\n", ocount);
	if (e->fd < 0) {
		errmsg(tag, "Tcreate: unable to create");
		return;
	}


	if (gr = getgrgid(dirgr))
		fprintf(stderr, "change to group: %s\n", gr->gr_name);

	if (fchown(e->fd, -1, dirgr)) {
		errmsg(tag, "Tcreate: fchown failed");
		perror("fchown");
		return;
	}

	e->open = 1;
	e->mode = mode;
	free(e->truename);
	e->truename = str;
#ifdef VERBOSE
	fprintf(stderr, "<full name = %s>", e->truename);
#endif
	strcpy(e->name, name);

	if (stat(e->truename, &stbuf)) {
		errmsg(tag, "Tcreate: stat failed");
		return;
	}
	e->qidpath = stbuf.st_ino;
	if (S_ISDIR(stbuf.st_mode)) {
		e->qidpath |= DIRMASK;
	} else {
		e->qidpath &= ~DIRMASK;
	}

	e->qidver = 0;

	putpacket("12244", Rcreate, tag, fid, e->qidpath, e->qidver);
}

void handleread(void) {
	fidelem *e;
	long long int offset;
	unsigned int count;
	struct stat stbuf;
	int i, j, l;
	char *str;
	DIR *d;
	struct dirent *dirbuf;
	int tag;

	getpacket("2282", &tag, &fid, &offset, &count);
#ifdef VERBOSE
	fprintf(stderr, "Request for %d bytes\n", count);
#endif

	GOODTAG(Tread);

	e = fidretrieve(fid);
	if (!e) {
		errmsg(tag, "Tread: no such fid");
		return;
	}

	if (e->childfd != -1) {
		errmsg(tag, "Tread: fid busy");
		return;
	}

	if (!e->open) {
		errmsg(tag, "Tread: file not open");
		return;
	}

	if (e->mode & 1) {
		errmsg(tag, "Tread: file not opened for reading");
		return;
	}

	stat(e->truename, &stbuf);

	if (S_ISDIR(stbuf.st_mode)) {

		if (offset % DIRLEN != 0) {
			errmsg(tag, "Tread: offset not 0 modulo DIRLEN for directory");
			return;
		}
		if ((count % DIRLEN) != 0) {
			errmsg(tag, "Tread: count not 0 modulo DIRLEN for directory");
			return;
		}

		d = opendir(e->truename);
		if (!d) {
			errmsg(tag, "Tread: can't open directory");
			return;
		}

		for (i=0; i<(offset / DIRLEN); i++) {
			dirbuf = readdir(d);
			if (!dirbuf) break;
		}

		l = strlen(e->truename);
		str = (char *) salloc(l + NAMELEN + 1);
		strcpy(str, e->truename);
		str[l] = '/';
		l++;
		j = count / DIRLEN;
		count = 0;
		for (i=0; i < j; i++) {
			dirbuf = readdir(d);
			if (!dirbuf) {
				break;
			} else {
				strcpy(&str[l], dirbuf->d_name);
				if (stat(str, &stbuf)) {
					fprintf(stderr, "Error: %s\n", dirbuf->d_name);
					perror("Tread: stat");
					//errmsg(tag, "Tread: stat failed");
					//return;
				}
				getstat(dirbuf->d_name, stbuf, &data[count]);
				count += DIRLEN;
			}
		}
		putpacket("12221s", Rread, tag, fid, count, '*', data, count);
		return;
	}

	if (offset > stbuf.st_size) {
		//count = 0;
		putpacket("12221", Rread, tag, fid, 0, '*');

	} else {
		pid_t p;
		int pfd[2];
		fd_set rfds;
		int maxfd;

		lseek(e->fd, offset, SEEK_SET);
				
		if (S_ISCHR(stbuf.st_mode) || S_ISBLK(stbuf.st_mode)) {
			if (pipe(pfd)) {
				perror("pipe");
				errmsg(tag, "pipe failed");
				return;
			}

			p = fork();

			if (p == -1) {
				perror("fork");
				close(pfd[0]);
				close(pfd[1]);
				errmsg(tag, "fork failed");
				return;

			} else if (p == 0) {
fprintf(stderr, "PIPE FD0 %d\n", pfd[0]);
				/* child */
				close(parentfd[0]);
				close(pfd[1]);
				if (e->fd > pfd[0])
					maxfd = e->fd + 1;
				else
					maxfd = pfd[0] + 1;

				FD_ZERO(&rfds);
				FD_SET(pfd[0], &rfds);
				FD_SET(e->fd, &rfds);
				if (select(maxfd, &rfds, NULL, NULL, NULL) == -1) {
					perror("select");
					strcpy(data, "select error");
					fputpacket(parentfd[1], "2212s", tag, 3 + ERRLEN,
							Rerror, tag, data, ERRLEN);
				}
				
				if (FD_ISSET(pfd[0], &rfds)) {
					/* parent is telling us something */
					fgetpacket(pfd[0], "s", data, 1);
					if (data[0] == Tflush) {
						/* Abort outstanding I/O */
						close(pfd[0]);
						exit(0);
					}
				}

				if (FD_ISSET(e->fd, &rfds)) {
					count = read(e->fd, data, count);
					if (count == -1) {
						perror("read");
						strcpy(data, "Tread: read error");
						fputpacket(parentfd[1], "2212s", tag, 3 + ERRLEN,
								Rerror, tag, data, ERRLEN);
					} else {
						fprintf(stderr, "<child: sending info to %d>\n", parentfd[1]);
						fputpacket(parentfd[1], "2212221s", tag, 8 + count,
								Rread, tag, fid, count, '*', data, count);
					}
					close(pfd[0]);
					exit(0);
				}
			} else {
				/* parent */
				close(pfd[0]);
				e->childfd = pfd[1];
				pendadd(pendnew(tag, fid, p));
			}
		} else {
			count = read(e->fd, data, count);
			if (count == -1) {
				perror("read");
				errmsg(tag, "Tread: read error");
				return;
			}
			putpacket("12221s", Rread, tag, fid, count, '*', data, count);
		}
	}
}

void handlewrite(void) {
	fidelem *e;
	off_t offset;
	unsigned int count;
	struct stat stbuf;
	int i;
	char pad;
	int tag;

	getpacket("22821", &tag, &fid, &offset, &count, &pad);

	GOODTAG(Twrite);

	if (count > MAXFDATA) {
		errmsg(tag, "Twrite: count is too large");
		return;
	}

#ifdef VERBOSE
	fprintf(stderr, "<writing %d bytes>", count);
#endif

	getpacket("s", data, count);

	e = fidretrieve(fid);
	if (!e) {
		errmsg(tag, "Twrite: no such fid");
		return;
	}

	if (e->childfd != -1) {
		errmsg(tag, "Twrite: fid busy");
		return;
	}

	if (!e->open) {
		errmsg(tag, "Twrite: file not open");
		return;
	}

	if (!((e->mode & 1) || (e->mode & 2))) {
		errmsg(tag, "Twrite: file not opened for writing");
		return;
	}

	stat(e->truename, &stbuf);

	if (S_ISDIR(stbuf.st_mode)) {
		errmsg(tag, "Twrite: cannot write to directories");
		return;
	}

	if (e->appendonly || offset > stbuf.st_size) {
		offset = stbuf.st_size;
	}

	{
		pid_t p;
		int pfd[2];
		fd_set rfds;
		int maxfd;

		lseek(e->fd, offset, SEEK_SET);
				
		if (S_ISCHR(stbuf.st_mode) || S_ISBLK(stbuf.st_mode)) {
			if (pipe(pfd)) {
				perror("pipe");
				errmsg(tag, "pipe failed");
				return;
			}

			p = fork();

			if (p == -1) {
				perror("fork");
				close(pfd[0]);
				close(pfd[1]);
				errmsg(tag, "fork failed");
				return;

			} else if (p == 0) {
fprintf(stderr, "PIPE FD0 %d\n", pfd[0]);
				/* child */
				close(parentfd[0]);
				close(pfd[1]);
				if (e->fd > pfd[0])
					maxfd = e->fd + 1;
				else
					maxfd = pfd[0] + 1;

				FD_ZERO(&rfds);
				FD_SET(pfd[0], &rfds);
				FD_SET(e->fd, &rfds);
				if (select(maxfd, &rfds, NULL, NULL, NULL) == -1) {
					perror("select");
					strcpy(data, "select error");
					fputpacket(parentfd[1], "2212s", tag, 3 + ERRLEN,
							Rerror, tag, data, ERRLEN);
				}
				
				if (FD_ISSET(pfd[0], &rfds)) {
					/* parent is telling us something */
					fgetpacket(pfd[0], "s", data, 1);
					if (data[0] == Tflush) {
						/* Abort outstanding I/O */
						close(pfd[0]);
						exit(0);
					}
				}

				if (FD_ISSET(e->fd, &rfds)) {
					count = write(e->fd, data, count);

					if (count == -1) {
						perror("write");
						strcpy(data, "Twrite: write error");
						fputpacket(parentfd[1], "2212s", tag, 3 + ERRLEN,
								Rerror, tag, data, ERRLEN);
					} else {
						fprintf(stderr, "<child: sending info to %d>\n", parentfd[1]);
						fputpacket(parentfd[1], "221222", tag, 7,
								Rwrite, tag, fid, count);
					}
					close(pfd[0]);
					exit(0);
				}
			} else {
				/* parent */
				close(pfd[0]);
				e->childfd = pfd[1];
				pendadd(pendnew(tag, fid, p));
			}
		} else {
			count = write(e->fd, data, count);
			if (count == -1) {
				perror("write");
				errmsg(tag, "Twrite: write error");
				return;
			}
			putpacket("1222", Rwrite, tag, fid, count);
		}
	}
}

void handleremove(void) {
	fidtype fid;
	fidelem *e;
	int tag;
	struct stat stbuf;

	getpacket("22", &tag, &fid);

	GOODTAG(Tremove);

	e = fiddelete(fid);

	if (!e) {
		errmsg(tag, "Tremove: No such fid");
		return;
	}

	if (e->childfd != -1) {
		errmsg(tag, "Tremove: fid busy");
		return;
	}

	if (e->open) {
		ocount--;
		fprintf(stderr, "CLOSING *****%d\n", ocount);
		if (close(e->fd)) {
			perror("close");
		}
	}

	if (stat(e->truename, &stbuf)) {
		perror("stat");
		errmsg(tag, "Tremove: stat failed");
		free(e);
		return;
	}

	if (S_ISDIR(stbuf.st_mode)) {
		if (rmdir(e->truename) == -1) {
			perror("rmdir");
			errmsg(tag, "Tremove: rmdir failed");
			free(e);
			return;
		}
	} else {
		if (unlink(e->truename) == -1) {
			perror("unlink");
			errmsg(tag, "Tremove: unlink failed");
			free(e);
			return;
		}
	}

	free(e);

	putpacket("122", Rremove, tag, fid);
}

void handlestat(void) {
	fidtype fid;
	fidelem *e;
	struct stat stbuf;
	char fileinfo[DIRLEN];
	int tag;

	getpacket("22", &tag, &fid);
#ifdef VERBOSE
	fprintf(stderr, "<Tstat received>\n");
#endif

	GOODTAG(Tstat);

	e = fidretrieve(fid);
	if (!e) {
		errmsg(tag, "Tstat: no such fid");
		return;
	}

	if (stat(e->truename, &stbuf)) {
		errmsg(tag, "Tstat: stat failed");
		return;
	}

#ifdef VERBOSE
	fprintf(stderr, "<stat on %s>", e->truename);
#endif
	getstat(e->name, stbuf, fileinfo);

	putpacket("122s", Rstat, tag, fid, fileinfo, DIRLEN);

}

void handlewstat(void) {
	fidelem *e;
	char name[NAMELEN + 1];
	char *str;
	int l;
	struct stat stbuf;
	fidtype fid;
	char fileinfo[DIRLEN];
	struct group *gr;
	int tag;
	struct utimbuf newtime;
	time_t mtime;

	getpacket("22s", &tag, &fid, fileinfo, DIRLEN);
#ifdef VERBOSE
	fprintf(stderr, "<Twstat received>\n");
#endif

	GOODTAG(Twstat);

	fileinfo[NAMELEN] = '\0';
	/* First part holds the name */
	strcpy(name, fileinfo);

	e = fidretrieve(fid);

	if (!e) {
		errmsg(tag, "Twstat: no such fid");
		return;
	}

	if (e->childfd != -1) {
		errmsg(tag, "Twrite: fid busy");
		return;
	}

	/* rename if name changed */
	if (strcmp(name, e->name)) {
		l = strlen(e->truename);

		str = (char *) salloc(l + NAMELEN + 1);
		while (l > 0 && e->truename[l] != '/') l--;
		l++;
		strcpy(str, e->truename);
		strcpy(&str[l], name);
		fprintf(stderr, "<Renaming to %s>\n", str);
		if (rename(e->truename, str) == -1) {
			errmsg(tag, "Twstat: rename failed");
			perror("rename");
			return;
		}
		free(e->truename);
		e->truename = str;
	}

	/* Change time */
	l = 3 * NAMELEN + 4 * 4;
	mtime = fileinfo[l++]; 
	mtime |= fileinfo[l++] << 8;
	mtime |= fileinfo[l++] << 16;
	mtime |= fileinfo[l++] << 24;
	newtime.actime = newtime.modtime = mtime;

	if (utime(e->truename, &newtime)) {
		errmsg(tag, "Twstat: utime failed");
		perror("utime");
	}

	/* Change group */
	fileinfo[3 * NAMELEN] = '\0';
	strcpy(name, &fileinfo[2 * NAMELEN]);
	if (gr = getgrnam(name)) {
		if (chown(e->truename, -1, gr->gr_gid)) {
			errmsg(tag, "Twstat: chown failed");
			perror("chown");
			return;
		}
	} else {
		errmsg(tag, "Twstat: getgrnam failed");
		perror("getgrnam");
		return;
	}

#ifdef VERBOSE
	fprintf(stderr, "<full name = %s>", e->truename);
#endif
	e->qidpath = qidpath(stbuf);
	e->qidver = qidver(stbuf);

	putpacket("122", Rwstat, tag, fid);
}

void handlewalk(void) {
	fidelem *e;
	char name[NAMELEN + 1];
	char *str;
	int l;
	struct stat stbuf;
	int tag;

	name[NAMELEN] = '\0';

	getpacket("22s", &tag, &fid, name, NAMELEN);

	GOODTAG(Twalk);

	e = fidretrieve(fid);
	if (!e) {
		errmsg(tag, "Twalk: no such fid");
		return;
	}

	if (stat(e->truename, &stbuf)) {
		errmsg(tag, "Twalk: unable to stat fid");
		return;
	}

	if (!S_ISDIR(stbuf.st_mode)) {
		errmsg(tag, "Twalk: fid is not a directory");
		return;
	}

	if (e->open) {
ocount--;
fprintf(stderr, "CLOSE ***** %d\n", ocount);
		if (close(e->fd)) {
			perror("close");
		}
		e->open = 0;
	}

	l = strlen(e->truename);

	if (!strcmp(name, "..")) {
		while ((l > 0) && (e->truename[l] != '/')) l--;
		if (l == 0) {
			str = (char *) salloc(2);
			strcpy(str, "/");
		} else {
			str = (char *) salloc(l);
			/* TODO: This is lazy! */
			e->truename[l] = '\0';
			strcpy(str, e->truename);
			e->truename[l] = '/';
		}
	} else if (!strcmp(name, ".")) {

		putpacket("12244", Rwalk, tag, fid, e->qidpath, e->qidver);
		return;

	} else if (!strcmp(name, "/")) {
		str = (char *) salloc(2);
		strcpy(str, "/");

	} else {
		/* TODO: More laziness! */
		if (!strcmp(e->truename, "/")) {
			l--;
		}

		str = (char *) salloc(l + NAMELEN + 2);
		strcpy(str, e->truename);
		str[l] = '/';
		l++;

		strcpy(&str[l], name);
	}

	if (stat(str, &stbuf)) {
		fprintf(stderr, "<truename = %s>", str);
		errmsg(tag, "Twalk: stat failed");
		perror("Twalkstat");
		return;
	}

	if (!strcmp(name, "..")) {
		if (!strcmp(str, "/"))
			strcpy(e->name, "/");
		else {
			l = strlen(e->name);
			while ((l > 0) && (str[l] != '/')) l--;
			strcpy(e->name, &str[l + 1]);
		}
	} else {
		strcpy(e->name, name);
	}

	free(e->truename);
	e->truename = str;
#ifdef VERBOSE
	fprintf(stderr, "<full name = %s>", e->truename);
#endif
	e->qidpath = qidpath(stbuf);
	e->qidver = qidver(stbuf);

	putpacket("12244", Rwalk, tag, fid, e->qidpath, e->qidver);
}

void handleclwalk(void) {
	fidelem *e;
	fidtype newfid;
	int tag;
	char name[NAMELEN + 1];
	char *str;
	int l;
	struct stat stbuf;

	getpacket("222s", &tag, &fid, &newfid, name, NAMELEN);

	GOODTAG(Tclwalk);

	e = fidretrieve(fid);
	if (!e) {
		errmsg(tag, "Tclwalk: No such fid");
		return;
	}

	if (fidretrieve(newfid)) {
		errmsg(tag, "Tclwalk: newfid already used");
		return;
	}

	e = fidnew(newfid, e->name, e->truename, e->qidpath, e->qidver);

#ifdef VERBOSE
	fprintf(stderr, "<cloned %d -> %d>", fid, newfid);
#endif

	name[NAMELEN] = '\0';

	if (stat(e->truename, &stbuf)) {
		errmsg(tag, "Tclwalk: unable to stat newfid");
		free(e);
		return;
	}

	if (!S_ISDIR(stbuf.st_mode)) {
		errmsg(tag, "Tclwalk: newfid is not a directory");
		free(e);
		return;
	}

	if (e->open) {
ocount--;
fprintf(stderr, "CLOSE ***** %d\n", ocount);
		if (close(e->fd)) {
			perror("close");
		}
		e->open = 0;
	}

	l = strlen(e->truename);

	if (!strcmp(name, "..")) {
		while ((l > 0) && (e->truename[l] != '/')) l--;
		if (l == 0) {
			str = (char *) salloc(2);
			strcpy(str, "/");
		} else {
			str = (char *) salloc(l);
			/* TODO: This is lazy! */
			e->truename[l] = '\0';
			strcpy(str, e->truename);
			e->truename[l] = '/';
		}
	} else if (!strcmp(name, ".")) {
		fidinsert(e);
		putpacket("12244", Rclwalk, tag, newfid, e->qidpath, e->qidver);
		return;

	} else if (!strcmp(name, "/")) {
		str = (char *) salloc(2);
		strcpy(str, "/");

	} else {
		/* TODO: More laziness! */
		if (!strcmp(e->truename, "/")) {
			l--;
		}

		str = (char *) salloc(l + NAMELEN + 2);
		strcpy(str, e->truename);
		str[l] = '/';
		l++;

		strcpy(&str[l], name);
	}

	if (stat(str, &stbuf)) {
		fprintf(stderr, "<truename = %s>", str);
		errmsg(tag, "Tclwalk: stat failed");
		perror("stat");
		return;
	}

	if (!strcmp(name, "..")) {
		if (!strcmp(str, "/"))
			strcpy(e->name, "/");
		else {
			l = strlen(e->name);
			while ((l > 0) && (str[l] != '/')) l--;
			strcpy(e->name, &str[l + 1]);
		}
	} else {
		strcpy(e->name, name);
	}

	free(e->truename);
	e->truename = str;
#ifdef VERBOSE
	fprintf(stderr, "<full name = %s>", e->truename);
#endif
	e->qidpath = qidpath(stbuf);
	e->qidver = qidver(stbuf);

	fidinsert(e);
	putpacket("12244", Rclwalk, tag, newfid, e->qidpath, e->qidver);
}
