/* File browser using the 9P protocol */
/* Ben Lin
 * 	10/4/98
 *		Started this file, copied from 9pc.c and 9ps.c
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdarg.h>
#include <netinet/in.h>
#include "unix9p.h"

#define EXPECT(x) if (ch != x) { errmsg(#x " expected\n"); close(sfd); exit(1); }
#define GETARG() sscanf(&s[1], "%s", name);

int sfd;

int putpacket(char *, ...);

/* Displays an error message */

void errmsg(char *s) {
	fprintf(stderr, "browser: %s\n", s);
}

int putpacket(char *fmt, ...) {
	va_list ap;
	static char packet[MAXFDATA + 16];
	unsigned char ch;
	char *s;
	long long int ln;
	int n;
	int i, j;

	va_start(ap, fmt);
	i = 0;
	while(*fmt) {
		switch(*fmt++) {
			case '1':	/* one byte */
				ch = va_arg(ap, char);
				packet[i++] = ch;
				break;

			case '2':
				n = va_arg(ap, int);
				packet[i++] = (char) n;	/* high-bits get truncated */
				packet[i++] = (char) (n >> 8);
				break;
			case '4':
				n = va_arg(ap, int);
				packet[i++] = (char) n;
				packet[i++] = (char) (n = n >> 8);
				packet[i++] = (char) (n = n >> 8);
				packet[i++] = (char) (n >> 8);
				break;
			case '8':
				ln = va_arg(ap, long int);
				packet[i++] = (char) ln;
				packet[i++] = (char) (ln = ln >> 8);
				packet[i++] = (char) (ln = ln >> 8);
				packet[i++] = (char) (ln = ln >> 8);
				packet[i++] = (char) (ln = ln >> 8);
				packet[i++] = (char) (ln = ln >> 8);
				packet[i++] = (char) (ln = ln >> 8);
				packet[i++] = (char) (ln >> 8);
				break;

			case 's':
				s = va_arg(ap, char *);
				n = va_arg(ap, int);
				for (j=0; j<n; j++) {
					packet[i++] = s[j];
				}
				break;

			default:
				errmsg("BUG! Bad char in format string");
				break;
		}
	}

	va_end(ap);

	write(sfd, packet, i);
	return(0);
}

int getpacket(char *fmt, ...) {
	va_list ap;
	static unsigned char packet[MAXFDATA + 16];
	int length;
	int i, j;
	int *ip;
	long long int *lp;
	char *cp;

	va_start(ap, fmt);
	length = 0;
	i = 0;
	while(fmt[i]) {
		switch(fmt[i++]) {
			case '1':	/* one byte */
				va_arg(ap, char);
				length++;
				break;

			case '2':
				va_arg(ap, char);
				length+=2;
				break;

			case '4':
				va_arg(ap, char);
				length+=4;
				break;

			case '8':
				va_arg(ap, char);
				length+=8;
				break;

			case 's':
				va_arg(ap, char);
				length+=va_arg(ap, int);
				break;

			default:
				errmsg("BUG! Bad char in format string");
				break;
		}
	}
	
	i = 0;
	while (length > 0) {
		j = read(sfd, &packet[i], length);
		if (j == -1) {
			errmsg("Error reading from port");
			close(sfd);
			exit(100);
		} else {
			i += j;
			length -= j;
		}
	};

	i = 0;
	va_start(ap, fmt);

	while(*fmt) {
		switch(*fmt++) {
			case '1':	/* one byte */
				cp = va_arg(ap, char *);
				*cp = packet[i++];
				break;

			case '2':
				ip = va_arg(ap, int *);
				*ip = packet[i++]; 
				*ip |= packet[i++] << 8;
				break;

			case '4':
				ip = va_arg(ap, int *);
				*ip = packet[i++]; 
				*ip |= packet[i++] << 8;
				*ip |= packet[i++] << 16;
				*ip |= packet[i++] << 24;
				break;

			case '8':
				lp = va_arg(ap, long long int *);
				*lp = packet[i++];
				for (j=1; j<8; j++) {
					*lp |= ((long long int) packet[i++]) << (8*j);
				}
				break;

			case 's':
				cp = va_arg(ap, char *);
				length = va_arg(ap, int);
				for (j=0; j<length; j++) {
					cp[j] = packet[i++];
				}

				break;

			default:
				fprintf(stderr, "9ps: BUG! Bad char in format string\n");
				break;
		}
	}
	va_end(ap);

	return(0);
}

/* convert to long integer */
long int stoli(unsigned char *s) {
	int i;
	unsigned long int li;
	
	li = s[0];
	for (i=1; i<4; i++) {
		li |= ((long int) s[i]) << (8 * i);
	}

	return(li);

}

/* MAIN */

main(int argc, char *argv[]) {
	unsigned char s[MAXFDATA];
	int i, count;
	int tag, fid;
    long long int qid;
	int perm;
	char ch, pad;
	long long int offset;
	struct sockaddr_in sock9P;
	char name[NAMELEN];
	char dirflag;
	
	sfd = socket(AF_INET, SOCK_STREAM, 0);

	if (sfd == -1) {
		perror("socket");
		exit(1);
	}

	sock9P.sin_family = AF_INET;
	sock9P.sin_port = htons(PORT_9P);
	sock9P.sin_addr.s_addr = inet_addr("127.0.0.1");
	bzero(&(sock9P.sin_zero), 8);

	if (connect(sfd, (struct sockaddr *)&sock9P, sizeof(struct sockaddr)) != 0) {
		perror("connect");
		exit(1);
	}

	fprintf(stderr, "Testing: sending Tnop...");
	putpacket("12", Tnop, NOTAG);
	getpacket("12", &ch, &tag);
	if (ch != Rnop) {
		fprintf(stderr, "Rnop expected\n");
		exit(1);
	}

	fprintf(stderr, "Rnop received\n");

	strcpy(s, "/");
	printf("login:");
	scanf("%s", name);

#ifdef INFERNO
	putpacket("122ss", Tattach, 123, 0, name, NAMELEN, s, NAMELEN);
	getpacket("1228", &ch, &tag, &fid, &qid);
#else
	putpacket("122ssss", Tattach, 123, 0, name, NAMELEN, s, NAMELEN,
			s, TICKETLEN, s, AUTHENTLEN);
	getpacket("1228s", &ch, &tag, &fid, &qid, s, AUTHENTLEN);
#endif

	EXPECT(Rattach);

	putpacket("1221", Topen, 13, 0, 0);
	getpacket("1228", &ch, &tag, &fid, &qid);
	EXPECT(Ropen);

	dirflag = 1;
	printf(">");
	scanf("%s", &s);
	while (!feof(stdin)) {
		switch(s[0]) {
			case 'w':
				if (!dirflag) {
					printf("Can only use w in a directory\n");
					break;
				}

				GETARG();
				printf("Walking to %s\n", name);
				putpacket("122s", Twalk, 31, 0, name, NAMELEN);
				getpacket("1228", &ch, &tag, &fid, &qid);
				EXPECT(Rwalk);
				printf("qid = %llX\n", qid);

				putpacket("122", Tstat, 1, 0);
				getpacket("122s", &ch, &tag, &fid, s, DIRLEN);
				EXPECT(Rstat);
				
				if (s[3*NAMELEN + 11] & 0x80) dirflag = 1;
				else dirflag = 0;

				putpacket("1221", Topen, 13, 0, 0);
				getpacket("1228", &ch, &tag, &fid, &qid);
				EXPECT(Ropen);

				break;

			case 'd':
				if (dirflag) {
					printf("Directory:\n");

					putpacket("12282", Tread,  42, 0, 0, DIRLEN*50);
					getpacket("12221", &ch, &tag, &fid, &count, &pad);
					EXPECT(Rread);
					getpacket("s", s, count);

					printf("%d bytes read (%d entries)\n", count, count / DIRLEN);
					for (i=0; i<(count / DIRLEN); i++) {
						printf("%d\t%s\n", stoli(&s[DIRLEN*i + 28*3 + 20]),
								&s[DIRLEN*i]);
					}
				} else {
					printf("Can only use d in a directory\n");
				}
				break;

			case 's':
				printf("Stats:\n");
				putpacket("122", Tstat, 1, 0);
				getpacket("122s", &ch, &tag, &fid, s, DIRLEN);
				EXPECT(Rstat);
				printf("Name: %s\n", &s[0]);
				printf("Owner: %s\n", &s[28]);
				printf("Group: %s\n", &s[28*2]);
				printf("qid: path = %u, ver = %u\n", stoli(&s[28*3]),
						stoli(&s[28*3+4]));
				printf("Length: %lu\n", stoli(&s[28*3+20]) +
						((long long int) stoli(&s[28*3+24]) << 32));
				break;

			case 'r':
				if (!dirflag) {
					putpacket("12282", Tread, 5, 0, 0, 100);
					getpacket("12221", &ch, &tag, &fid, &count, &pad);
					EXPECT(Rread);
					
					printf("Reading %d bytes\n", count);
					getpacket("s", s, count);
					for (i=0; i<count; i++) {
						putchar(s[i]);
					}
				} else {
					printf("Can only read files\n");
				}
				break;

			case 'C':
				if (dirflag) {
					GETARG();
					perm = 0765;
					putpacket("122s41", Tcreate, 987, 0, name, NAMELEN,
							perm, 2);
					getpacket("1228", &ch, &tag, &fid, &qid);
					EXPECT(Rcreate);
					dirflag = 0;
				} else {
					printf("Can only create in directories\n");
				}
				break;

			case 'D':
				if (dirflag) {
					GETARG();
					perm = 0777 | DIRMASK;
					putpacket("122s41", Tcreate, 987, 0, name, NAMELEN,
							perm, 2);
					getpacket("1228", &ch, &tag, &fid, &qid);
					EXPECT(Rcreate);
				} else {
					printf("Can only create in directories\n");
				}
				break;
			
			case 'W':
				if (!dirflag) {
					printf("Offset: ");
					scanf("%lld", &offset);
					printf("Enter data: ");
					scanf("%s", s);
					printf("Received %s (length = %d)\n", s, count = strlen(s));
					putpacket("122821s", Twrite, 1234, 0, 0, 8,
							'*', "12345678", 8);
					getpacket("1222", &ch, &tag, &fid, &count);
					EXPECT(Rwrite);
					printf("%d bytes written\n", count);

				} else {
					printf("Can only write to files\n");
				}
				break;

			case 'R':
				if (!dirflag) {
					putpacket("122", Tremove, 0, 0);
					getpacket("122", &ch, &tag, &fid);
					EXPECT(Rremove);
					printf("Quit the program now...");
				} else {
					printf("Can only remove files\n");
				}
				break;

			default:
				printf("unknown command\n");
				break;
		}

		printf(">");
		scanf("%s", &s);
	}
	putpacket("122", Tclunk, 153, 0);
	getpacket("122", &ch, &tag, &fid);
	EXPECT(Rclunk);

	putpacket("1", Tmax);
	close(sfd);

	return(0);
}
